#ifndef MSP_GL_PROGRAMPARSER_H_
#define MSP_GL_PROGRAMPARSER_H_

#include <deque>
#include <map>
#include <set>
#include <string>
#include <msp/io/base.h>
#include "programsyntax.h"

namespace Msp {
namespace GL {

class ProgramParser
{
private:
	enum OperatorType
	{
		NO_OPERATOR,
		BINARY,
		PREFIX,
		POSTFIX
	};

	enum Associativity
	{
		LEFT_TO_RIGHT,
		RIGHT_TO_LEFT
	};

	struct Operator
	{
		const char token[4];
		unsigned precedence;
		OperatorType type;
		Associativity assoc;
	};

	std::string source;
	std::string source_name;
	unsigned current_line;
	std::string::const_iterator iter;
	std::string::const_iterator source_end;
	bool allow_preprocess;
	bool allow_stage_change;
	std::string last_token;
	std::deque<std::string> next_tokens;
	ProgramSyntax::Module *module;
	ProgramSyntax::Stage *cur_stage;
	std::set<std::string> declared_types;

	static Operator operators[];

public:
	ProgramParser();
	~ProgramParser();

	ProgramSyntax::Module &parse(const std::string &, const std::string &);
	ProgramSyntax::Module &parse(IO::Base &, const std::string &);

private:
	void parse_source();

	std::string format_error(const std::string &);
	std::string format_syntax_error(const std::string &);

	const std::string &peek_token(unsigned = 0);
	const std::string &parse_token();
	std::string parse_token_();
	std::string parse_identifier();
	std::string parse_number();
	std::string parse_other();
	void skip_comment_and_whitespace();
	void expect(const std::string &);
	std::string expect_type();
	std::string expect_identifier();
	bool check(const std::string &);

	static bool is_interface_qualifier(const std::string &);
	static bool is_sampling_qualifier(const std::string &);
	static bool is_precision_qualifier(const std::string &);
	static bool is_qualifier(const std::string &);
	static bool is_builtin_type(const std::string &);
	bool is_type(const std::string &);
	bool is_identifier(const std::string &);

	void preprocess();
	void preprocess_pragma();
	void preprocess_pragma_msp();
	void preprocess_stage();

	RefPtr<ProgramSyntax::Node> parse_global_declaration();
	RefPtr<ProgramSyntax::Node> parse_statement();
	RefPtr<ProgramSyntax::Import> parse_import();
	RefPtr<ProgramSyntax::Precision> parse_precision();
	RefPtr<ProgramSyntax::Layout> parse_layout();
	void parse_block(ProgramSyntax::Block &, bool);
	RefPtr<ProgramSyntax::Expression> parse_expression(unsigned = 0);
	RefPtr<ProgramSyntax::BinaryExpression> parse_binary(const RefPtr<ProgramSyntax::Expression> &, const Operator *);
	RefPtr<ProgramSyntax::FunctionCall> parse_function_call(const ProgramSyntax::VariableReference &);
	RefPtr<ProgramSyntax::StructDeclaration> parse_struct_declaration();
	RefPtr<ProgramSyntax::VariableDeclaration> parse_variable_declaration();
	RefPtr<ProgramSyntax::FunctionDeclaration> parse_function_declaration();
	RefPtr<ProgramSyntax::InterfaceBlock> parse_interface_block();
	RefPtr<ProgramSyntax::Conditional> parse_conditional();
	RefPtr<ProgramSyntax::Iteration> parse_iteration();
	RefPtr<ProgramSyntax::Passthrough> parse_passthrough();
	RefPtr<ProgramSyntax::Return> parse_return();
};

} // namespace GL
} // namespace Msp

#endif
